# (C) Datadog, Inc. 2020-present
# All rights reserved
# Licensed under the Apache license (see LICENSE)
import importlib
from typing import Any

import click
from ._exceptions import MkDocsClickException


def load_command(module: str, attribute: str) -> click.BaseCommand:
    """
    Load and return the Click command object located at '<module>:<attribute>'.
    """
    command = _load_obj(module, attribute)

    if not isinstance(command, click.BaseCommand):
        raise MkDocsClickException(f"{attribute!r} must be a 'click.BaseCommand' object, got {type(command)}")

    return command


def _load_obj(module: str, attribute: str) -> Any:
    try:
        mod = importlib.import_module(module)
    except SystemExit:
        raise MkDocsClickException("the module appeared to call sys.exit()")  # pragma: no cover

    try:
        return getattr(mod, attribute)
    except AttributeError:
        raise MkDocsClickException(f"Module {module!r} has no attribute {attribute!r}")



# (C) Datadog, Inc. 2020-present
# All rights reserved
# Licensed under the Apache license (see LICENSE)
from .__version__ import __version__
from ._exceptions import MkDocsClickException
from ._extension import MKClickExtension, makeExtension

__all__ = ["__version__", "MKClickExtension", "MkDocsClickException", "makeExtension"]



# (C) Datadog, Inc. 2020-present
# All rights reserved
# Licensed under the Apache license (see LICENSE)
import click

NOT_A_COMMAND = "not-a-command"


@click.command()
@click.option("--count", default=1, help="Number of greetings.")
@click.option("--name", prompt="Your name", help="The person to greet.")
@click.option("--hidden", hidden=True)
def hello(count, name, hidden):
    """Simple program that greets NAME for a total of COUNT times."""


@click.group()
def cli():
    """Main entrypoint for this dummy program"""


@click.group(name="cli")
def cli_named():
    """Main entrypoint for this dummy program"""


@click.command()
def foo():  # No description
    pass  # pragma: no cover


@click.group()
def bar():
    """The bar command"""


@click.command(hidden=True)
def hidden():
    """The hidden command"""


bar.add_command(hello)

cli.add_command(foo)
cli.add_command(bar)
cli.add_command(hidden)

cli_named.add_command(foo)
cli_named.add_command(bar)
cli_named.add_command(hidden)


class MultiCLI(click.MultiCommand):
    def list_commands(self, ctx):
        return ["foo", "bar"]

    def get_command(self, ctx, name):
        cmds = {"foo": foo, "bar": bar}
        return cmds.get(name, None)


multi_named = MultiCLI(name="multi", help="Main entrypoint for this dummy program")
multi = MultiCLI(help="Main entrypoint for this dummy program")



# (C) Datadog, Inc. 2020-present
# All rights reserved
# Licensed under the Apache license (see LICENSE)
from contextlib import nullcontext

import pytest

from mkdocs_click._exceptions import MkDocsClickException
from mkdocs_click._loader import load_command


@pytest.mark.parametrize(
    "module, command, exc",
    [
        pytest.param("tests.app.cli", "cli", None, id="ok"),
        pytest.param("tests.app.cli", "doesnotexist", MkDocsClickException, id="command-does-not-exist"),
        pytest.param("doesnotexist", "cli", ImportError, id="module-does-not-exist"),
        pytest.param("tests.app.cli", "NOT_A_COMMAND", MkDocsClickException, id="not-a-command"),
    ],
)
def test_load_command(module: str, command: str, exc):
    with pytest.raises(exc) if exc is not None else nullcontext():  # type: ignore
        load_command(module, command)



# (C) Datadog, Inc. 2020-present
# All rights reserved
# Licensed under the Apache license (see LICENSE)
from mkdocs_click._processing import replace_blocks


def test_replace_options():
    """Replace a block with options."""

    source = """
# Some content
foo
::: target
    :option1: value1
    :optiøn2: value2
\t:option3:
    :option4:\x20
bar
""".strip()

    expected = """
# Some content
foo
{'option1': 'value1', 'optiøn2': 'value2', 'option3': '', 'option4': ''}
bar
""".strip()

    output = list(replace_blocks(source.splitlines(), title="target", replace=lambda **options: [str(options)]))
    assert output == expected.splitlines()


def test_replace_no_options():
    """Replace a block that has no options."""

    source = """
# Some content
foo
::: target
bar
""".strip()

    expected = """
# Some content
foo
> mock
bar
""".strip()

    output = list(replace_blocks(source.splitlines(), title="target", replace=lambda **options: ["> mock"]))
    assert output == expected.splitlines()


def test_other_blocks_unchanged():
    """Blocks other than the target block are left unchanged."""

    source = """
# Some content
::: target
::: plugin1
    :option1: value1
::: target
    :option: value
::: plugin2
    :option2: value2
bar
""".strip()

    expected = """
# Some content
::: plugin1
    :option1: value1
::: plugin2
    :option2: value2
bar
""".strip()

    output = list(replace_blocks(source.splitlines(), title="target", replace=lambda **kwargs: []))
    assert output == expected.splitlines()



# (C) Datadog, Inc. 2020-present
# All rights reserved
# Licensed under the Apache license (see LICENSE)
import re
from typing import Callable, Iterable, Iterator


def replace_blocks(lines: Iterable[str], title: str, replace: Callable[..., Iterable[str]]) -> Iterator[str]:
    """
    Find blocks of lines in the form of:

    ::: <title>
        :<key1>: <value>
        :<key2>:
        ...

    And replace them with the lines returned by `replace(key1="<value1>", key2="", ...)`.
    """

    options = {}
    in_block_section = False

    for line in lines:
        if in_block_section:
            match = re.search(r"^\s+:(?P<key>.+):(?:\s+(?P<value>\S+))?", line)
            if match is not None:
                # New ':key:' or ':key: value' line, ingest it.
                key = match.group("key")
                value = match.group("value") or ""
                if value.lower() in ["true", "false"]:
                    value = True if value.lower() == "true" else False
                options[key] = value
                continue

            # Block is finished, flush it.
            in_block_section = False
            yield from replace(**options)
            yield line
            continue

        match = re.search(rf"^::: {title}", line)
        if match is not None:
            # Block header, ingest it.
            in_block_section = True
            options = {}
        else:
            yield line



# (C) Datadog, Inc. 2020-present
# All rights reserved
# Licensed under the Apache license (see LICENSE)
from pathlib import Path
from textwrap import dedent

import pytest
from markdown import Markdown

import mkdocs_click

EXPECTED = (Path(__file__).parent / "app" / "expected.md").read_text()
EXPECTED_ENHANCED = (Path(__file__).parent / "app" / "expected-enhanced.md").read_text()
EXPECTED_SUB = (Path(__file__).parent / "app" / "expected-sub.md").read_text()
EXPECTED_SUB_ENHANCED = (Path(__file__).parent / "app" / "expected-sub-enhanced.md").read_text()


@pytest.mark.parametrize(
    "command, expected_name",
    [
        pytest.param("cli", "cli", id="cli-simple"),
        pytest.param("cli_named", "cli", id="cli-explicit-name"),
        pytest.param("multi_named", "multi", id="multi-explicit-name"),
        pytest.param("multi", "multi", id="no-name"),
    ],
)
def test_extension(command, expected_name):
    """
    Markdown output for a relatively complex Click application is correct.
    """
    md = Markdown(extensions=[mkdocs_click.makeExtension()])

    source = dedent(
        f"""
        ::: mkdocs-click
            :module: tests.app.cli
            :command: {command}
        """
    )

    expected = EXPECTED.replace("cli", expected_name)

    assert md.convert(source) == md.convert(expected)


def test_prog_name():
    """
    The :prog_name: attribute determines the name to display for the command.
    """
    md = Markdown(extensions=[mkdocs_click.makeExtension()])

    source = dedent(
        """
        ::: mkdocs-click
            :module: tests.app.cli
            :command: cli
            :prog_name: custom
        """
    )

    expected = EXPECTED.replace("cli", "custom")

    assert md.convert(source) == md.convert(expected)


def test_depth():
    """
    The :depth: attribute increases the level of headers.
    """
    md = Markdown(extensions=[mkdocs_click.makeExtension()])

    source = dedent(
        """
        # CLI Reference

        ::: mkdocs-click
            :module: tests.app.cli
            :command: cli
            :depth: 1
        """
    )

    expected = f"# CLI Reference\n\n{EXPECTED.replace('# ', '## ')}"

    assert md.convert(source) == md.convert(expected)


@pytest.mark.parametrize("option", ["module", "command"])
def test_required_options(option):
    """
    The module and command options are required.
    """
    md = Markdown(extensions=[mkdocs_click.makeExtension()])

    source = dedent(
        """
        ::: mkdocs-click
            :module: tests.app.cli
            :command: cli
        """
    )

    source = source.replace(f":{option}:", ":somethingelse:")

    with pytest.raises(mkdocs_click.MkDocsClickException):
        md.convert(source)


def test_enhanced_titles():
    """
    If `attr_list` extension is registered, section titles are enhanced with full command paths.

    See: https://github.com/DataDog/mkdocs-click/issues/35
    """
    md = Markdown(extensions=["attr_list"])
    # Register our extension as a second step, so that we see `attr_list`.
    # This is what MkDocs does, so there's no hidden usage constraint here.
    md.registerExtensions([mkdocs_click.makeExtension()], {})

    source = dedent(
        """
        ::: mkdocs-click
            :module: tests.app.cli
            :command: cli
        """
    )

    assert md.convert(source) == md.convert(EXPECTED_ENHANCED)


@pytest.mark.parametrize(
    "command, expected_name",
    [
        pytest.param("cli", "cli", id="cli-simple"),
        pytest.param("cli_named", "cli", id="cli-explicit-name"),
        pytest.param("multi_named", "multi", id="multi-explicit-name"),
        pytest.param("multi", "multi", id="no-name"),
    ],
)
def test_extension_with_subcommand(command, expected_name):
    """
    Markdown output for a relatively complex Click application is correct.
    """
    md = Markdown(extensions=[mkdocs_click.makeExtension()])

    source = dedent(
        f"""
        ::: mkdocs-click
            :module: tests.app.cli
            :command: {command}
            :list_subcommands: True
        """
    )

    expected = EXPECTED_SUB.replace("cli", expected_name)

    assert md.convert(source) == md.convert(expected)


@pytest.mark.parametrize(
    "command, expected_name",
    [
        pytest.param("cli", "cli", id="cli-simple"),
        pytest.param("cli_named", "cli", id="cli-explicit-name"),
        pytest.param("multi_named", "multi", id="multi-explicit-name"),
        pytest.param("multi", "multi", id="no-name"),
    ],
)
def test_enhanced_titles_with_subcommand(command, expected_name):
    """
    Markdown output for a relatively complex Click application is correct.
    """
    md = Markdown(extensions=["attr_list"])
    # Register our extension as a second step, so that we see `attr_list`.
    # This is what MkDocs does, so there's no hidden usage constraint here.
    md.registerExtensions([mkdocs_click.makeExtension()], {})

    source = dedent(
        f"""
        ::: mkdocs-click
            :module: tests.app.cli
            :command: {command}
            :list_subcommands: True
        """
    )

    expected = EXPECTED_SUB_ENHANCED.replace("cli", expected_name)

    assert md.convert(source) == md.convert(expected)



# (C) Datadog, Inc. 2020-present
# All rights reserved
# Licensed under the Apache license (see LICENSE)
from textwrap import dedent
from typing import cast

import click
import pytest

from mkdocs_click._docs import make_command_docs, _show_options
from mkdocs_click._exceptions import MkDocsClickException


@click.command()
@click.option("-d", "--debug", help="Include debug output")
def hello():
    """Hello, world!"""


@click.command()
@click.option("-d", "--debug", help="Include debug output")
def hello_escape_marker():
    """
    \b
    Hello, world!
    """


@click.command()
@click.option("-d", "--debug", help="Include debug output")
def hello_ascii_art():
    """
    \b
      ______  __       __    ______  __  ___
     /      ||  |     |  |  /      ||  |/  /
    |  ,----'|  |     |  | |  ,----'|  '  /
    |  |     |  |     |  | |  |     |    <
    |  `----.|  `----.|  | |  `----.|  .  \\
     \\______||_______||__|  \\______||__|\\__\\

    Hello, world!
    """


HELLO_EXPECTED = dedent(
    """
    # hello

    Hello, world!

    **Usage:**

    ```
    hello [OPTIONS]
    ```

    **Options:**

    ```
      -d, --debug TEXT  Include debug output
      --help            Show this message and exit.
    ```
    """
).lstrip()


def test_basic():
    output = "\n".join(make_command_docs("hello", hello))
    assert output == HELLO_EXPECTED


def test_depth():
    output = "\n".join(make_command_docs("hello", hello, depth=2))
    assert output == HELLO_EXPECTED.replace("# ", "### ")


def test_prog_name():
    output = "\n".join(make_command_docs("hello-world", hello))
    assert output == HELLO_EXPECTED.replace("hello", "hello-world")


def test_style_invalid():
    with pytest.raises(
        MkDocsClickException, match="invalid is not a valid option style, which must be either `plain` or `table`."
    ):
        list(make_command_docs("hello", hello, style="invalid"))


def test_basic_escape_marker():
    output = "\n".join(make_command_docs("hello", hello_escape_marker))
    assert output == HELLO_EXPECTED


def test_basic_ascii_art():
    output = "\n".join(make_command_docs("hello", hello_ascii_art, remove_ascii_art=True))
    assert output == HELLO_EXPECTED


@click.command()
@click.option("-d", "--debug", help="Include debug output")
@click.option("--choice", type=click.Choice(["foo", "bar"]), default="foo")
@click.option("--date", type=click.DateTime(["%Y-%m-%d"]))
@click.option("--range-a", type=click.FloatRange(0, 1), default=0)
@click.option("--range-b", type=click.FloatRange(0))
@click.option("--range-c", type=click.FloatRange(None, 1), default=0)
@click.option("--flag/--no-flag")
def hello_full():
    """Hello, world!"""


HELLO_FULL_TABLE_EXPECTED = dedent(
    """
    # hello

    Hello, world!

    **Usage:**

    ```
    hello [OPTIONS]
    ```

    **Options:**

    | Name | Type | Description | Default |
    | ---- | ---- | ----------- | ------- |
    | `-d`, `--debug` | text | Include debug output | None |
    | `--choice` | choice (`foo` &#x7C; `bar`) | N/A | `foo` |
    | `--date` | datetime (`%Y-%m-%d`) | N/A | None |
    | `--range-a` | float range (between `0` and `1`) | N/A | `0` |
    | `--range-b` | float range (`0` and above) | N/A | None |
    | `--range-c` | float range (`1` and below) | N/A | `0` |
    | `--flag` / `--no-flag` | boolean | N/A | `False` |
    | `--help` | boolean | Show this message and exit. | `False` |
    """
).lstrip()


@click.command()
def hello_minimal():
    """Hello, world!"""


HELLO_MINIMAL_TABLE_EXPECTED = dedent(
    """
    # hello

    Hello, world!

    **Usage:**

    ```
    hello [OPTIONS]
    ```

    **Options:**

    | Name | Type | Description | Default |
    | ---- | ---- | ----------- | ------- |
    | `--help` | boolean | Show this message and exit. | `False` |
    """
).lstrip()


@pytest.mark.parametrize(
    "command, expected",
    [
        pytest.param(hello_full, HELLO_FULL_TABLE_EXPECTED, id="full"),
        pytest.param(hello_minimal, HELLO_MINIMAL_TABLE_EXPECTED, id="minimal"),
    ],
)
def test_style_table(command, expected):
    output = "\n".join(make_command_docs("hello", command, style="table"))
    assert output == expected


class MultiCLI(click.MultiCommand):
    def list_commands(self, ctx):
        return ["single-command"]

    def get_command(self, ctx, name):
        return hello


@pytest.mark.parametrize(
    "multi",
    [
        pytest.param(MultiCLI("multi", help="Multi help"), id="explicit-name"),
        pytest.param(MultiCLI(help="Multi help"), id="no-name"),
    ],
)
def test_custom_multicommand(multi):
    """
    Custom `MultiCommand` objects are supported (i.e. not just `Group` multi-commands).
    """
    expected = dedent(
        """
        # multi

        Multi help

        **Usage:**

        ```
        multi [OPTIONS] COMMAND [ARGS]...
        ```

        **Options:**

        ```
          --help  Show this message and exit.
        ```

        ## hello

        Hello, world!

        **Usage:**

        ```
        multi hello [OPTIONS]
        ```

        **Options:**

        ```
          -d, --debug TEXT  Include debug output
          --help            Show this message and exit.
        ```
        """
    ).lstrip()

    output = "\n".join(make_command_docs("multi", multi))
    assert output == expected


@pytest.mark.parametrize(
    "multi",
    [
        pytest.param(MultiCLI("multi", help="Multi help"), id="explicit-name"),
        pytest.param(MultiCLI(help="Multi help"), id="no-name"),
    ],
)
def test_custom_multicommand_with_list_subcommands(multi):
    """
    Custom `MultiCommand` objects are supported (i.e. not just `Group` multi-commands).
    """
    expected = dedent(
        """
        # multi

        Multi help

        **Usage:**

        ```
        multi [OPTIONS] COMMAND [ARGS]...
        ```

        **Options:**

        ```
          --help  Show this message and exit.
        ```

        **Subcommands**

        - *hello*: Hello, world!

        ## hello

        Hello, world!

        **Usage:**

        ```
        multi hello [OPTIONS]
        ```

        **Options:**

        ```
          -d, --debug TEXT  Include debug output
          --help            Show this message and exit.
        ```
        """
    ).lstrip()

    output = "\n".join(make_command_docs("multi", multi, list_subcommands=True))
    assert output == expected


@pytest.mark.parametrize("show_hidden", [True, False])
@pytest.mark.parametrize("style", ["plain", "table"])
def test_show_hidden_option(show_hidden, style):
    @click.command()
    @click.option("--hidden", hidden=True)
    def _test_cmd(hidden):
        """Test cmd."""

    output = "\n".join(make_command_docs("_test_cmd", _test_cmd, style=style, show_hidden=show_hidden))
    assert ("--hidden" in output) == show_hidden


@pytest.mark.parametrize("show_hidden", [True, False])
def test_show_hidden_command(show_hidden):
    @click.command(hidden=True)
    def _test_cmd():
        """Test cmd."""

    output = "\n".join(make_command_docs("_test_cmd", _test_cmd, show_hidden=show_hidden))
    assert (output != "") == show_hidden


@pytest.mark.parametrize("show_hidden", [True, False])
def test_show_hidden_group(show_hidden):
    @click.group(hidden=True)
    def _test_group():
        """Test group."""

    @_test_group.command(hidden=False)
    def _test_cmd():
        """Test cmd."""

    output = "\n".join(make_command_docs("_test_group", _test_group, show_hidden=show_hidden))
    assert (output != "") == show_hidden


def test_show_options():
    @click.command()
    @click.option("--hidden", hidden=True)
    @click.option("--normal", hidden=False)
    def _test_cmd(hidden, normal):
        """Test cmd."""

    ctx = click.Context(cast(click.Command, _test_cmd), info_name="_test_cmd")
    opt_hidden = cast(click.Option, ctx.command.params[0])
    opt_normal = cast(click.Option, ctx.command.params[1])

    assert opt_hidden.hidden
    assert not opt_normal.hidden

    with _show_options(ctx):
        assert not opt_hidden.hidden
        assert not opt_normal.hidden

    assert opt_hidden.hidden
    assert not opt_normal.hidden



# (C) Datadog, Inc. 2020-present
# All rights reserved
# Licensed under the Apache license (see LICENSE)
from typing import Any, Iterator, List

from markdown.extensions import Extension
from markdown.extensions.attr_list import AttrListExtension
from markdown.preprocessors import Preprocessor

from ._docs import make_command_docs
from ._exceptions import MkDocsClickException
from ._loader import load_command
from ._processing import replace_blocks


def replace_command_docs(has_attr_list: bool = False, **options: Any) -> Iterator[str]:
    for option in ("module", "command"):
        if option not in options:
            raise MkDocsClickException(f"Option {option!r} is required")

    module = options["module"]
    command = options["command"]
    prog_name = options.get("prog_name", None)
    depth = int(options.get("depth", 0))
    style = options.get("style", "plain")
    remove_ascii_art = options.get("remove_ascii_art", False)
    show_hidden = options.get("show_hidden", False)
    list_subcommands = options.get("list_subcommands", False)

    command_obj = load_command(module, command)

    prog_name = prog_name or command_obj.name or command

    return make_command_docs(
        prog_name=prog_name,
        command=command_obj,
        depth=depth,
        style=style,
        remove_ascii_art=remove_ascii_art,
        show_hidden=show_hidden,
        list_subcommands=list_subcommands,
        has_attr_list=has_attr_list,
    )


class ClickProcessor(Preprocessor):
    def __init__(self, md: Any) -> None:
        super().__init__(md)
        self._has_attr_list = any(isinstance(ext, AttrListExtension) for ext in md.registeredExtensions)

    def run(self, lines: List[str]) -> List[str]:
        return list(
            replace_blocks(
                lines,
                title="mkdocs-click",
                replace=lambda **options: replace_command_docs(has_attr_list=self._has_attr_list, **options),
            )
        )


class MKClickExtension(Extension):
    """
    Replace blocks like the following:

    ::: mkdocs-click
        :module: example.main
        :command: cli

    by Markdown documentation generated from the specified Click application.
    """

    def extendMarkdown(self, md: Any) -> None:
        md.registerExtension(self)
        processor = ClickProcessor(md)
        md.preprocessors.register(processor, "mk_click", 141)


def makeExtension() -> Extension:
    return MKClickExtension()
